/*
    Copyright 2010 Luis Miguel Merino Bonilla

    This file is part of GNU Taste

    GNU Taste is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GNU Taste is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with GNU Taste. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __HASH__
#define __HASH__

#include "feature_vector.h"
#include <cmath>
#include <iostream>

#define hash(x,size) x % size

using namespace std;

class Serializable {
	virtual void toStream(ostream &stream) = 0;
	virtual void fromStream(istream &stream) = 0;
	friend ostream &operator<<(ostream &stream, Serializable & ob){
		ob.toStream(stream);
		return stream;
	}

	friend istream & operator>>(istream &stream, Serializable & ob){
		ob.fromStream(stream);
		return stream;
	}
};

template <class T>
class Hash : public Serializable {
	//Clase auxiliar para almacenar los pares key/value enlazados
	class Par{
	public:
		int key;
		T & value;
		Par * next;
		Par(int k, T &v, Par * n) : key(k) , value(v), next(n) {}
	};

	//Clase para almacenar las entradas de la tabla hash
	class Entry{
	public:
		int k;
		Par * next;
		Entry() : k(0), next(0) {}
	};
		

private:
	Entry * array;
	int size;
	int entries;

public:
	//Debemos reimplementar el iterador para evitar el uso de set/keys
	//y recorrer directamente la tabla por las casillas ocupadas

	class Iterator{
	private:
		Par * p;
		Hash<T> & _hash;
	public:
		int position;

		Iterator(Hash<T> * h) : _hash(*h) {
			//Inicializamos el iterador a la primera casilla
			position = 0;
			//Avanzamos mientras la celda que recorremos no tenga entradas
			while(_hash.array[position].next == 0 && position<_hash.size)
				++position;
			//Devolvemos el primer par
			p = _hash.array[position].next;
		}
		void operator++(){
			// ¿Cómo buscamos la siguiente entrada?

			// Si tenemos una casilla detrás del puntero, avanzamos
			if(p->next)
				p = p->next;
			else{
				//Si no, tenemos que buscar la siguiente entrada con items
				while(position<_hash.size){
					//cout << position << "." << p->key << endl;
					++position;
					if(_hash.array[position].next != 0)
						break;

				}
				//Ya lo tenemos, devolvemos la primera celda
				if(position<_hash.size)
					p = _hash.array[position].next;
			}
			//cout << position << ":" << p->key << endl;
		}
		bool end(){ 
			// ¿Cuando hemos llegado al final?
			// 	* p->next apunta a 0
			// 	* position == size
			return position == _hash.size && p->next == 0;
		}
		Par * operator->(){ return p; }
		T & operator*(){ 
			return p->value;
		}
	};

	Hash(int max_size = 2500) : size(max_size) , entries(0) {
		array = new Entry [size];
	};
	
	//Si modificamos T&v por T v, segmentation fault :-?
	void insert(int k, T &v){
		int h = hash(k, size);
		Par *p = new Par(k,v,array[h].next);
		array[h].k = k;
		array[h].next = p;
		++entries;
		//cout << "Insertando " <<k<<":"<<v<<endl;
	}


	T & operator[](int k){
		int h = hash(k,size);
		Par * p = array[h].next;
		while(p != 0){
			if(p->key == k)
				return p->value;
			p = p->next;
		}

		T * new_item = new T();
		insert(k, *new_item);
		return *new_item;
	}

	Par & get_pair(int k){
		int h = hash(k, size);
		Par * p = array[h].next;
		while(p != 0){
			if(p->key == k){
				return *p;
			}
			p = p->next;
		}
	}

	T * get(int k){
		int h = hash(k,size);
		Par * p = array[h].next;
		while(p != 0){
			if(p->key == k){
				return &(p->value);
			}
			p = p->next;
		}
		return 0;
	}

	Iterator & begin(){
		Iterator * it = new Iterator(this);
		return *(it);
	}

	void toStream(ostream &stream){
		stream << size << endl;
		stream << entries << endl;
		for(int i=0; i<size; ++i){
			Par *p = array[i].next;
			while(p != 0){
				stream << p->key << " " << p->value << " ";
				p = p->next;
			}
			stream << endl;
		}
	}

	void fromStream(istream &stream){
		int numEntries;		
		stream >> this->size;
		stream >> numEntries;
		int key;
		T value;
		for(int i=0; i<numEntries; ++i){
			stream >> key >> value;
			this->insert(key,*(new T(value)));
//			cout << "("<<i<<"/"<<entries<<")"<<"key: " << key << ". " << value << endl;
		}	
	}	

	friend istream &operator>>(istream &stream, Hash<T> & ob){
		stream >> ob.size;
		return stream;
	}
};

#endif
